from multiprocessing import BoundedSemaphore
from typing import List, Optional

from skyplane.compute.ibmcloud.ibmcloud_auth import IBMCloudAuthentication
from skyplane.compute.ibmcloud.ibmcloud_server import IBMCloudServer
from skyplane.compute.ibmcloud.ibm_gen2.vpc_backend import IBMVPCBackend
from skyplane.compute.ibmcloud.ibm_gen2.config import load_config, REGIONS

from skyplane.compute.cloud_provider import CloudProvider
from skyplane.utils import imports


class IBMCloudProvider(CloudProvider):
    def __init__(
        self,
        key_prefix: str = "skyplane",
        auth: Optional[IBMCloudAuthentication] = None,
    ):
        super().__init__()
        self.key_prefix = key_prefix
        self.auth = auth if auth else IBMCloudAuthentication()
        self.regions_vpc = {}
        self.provisioning_semaphore = BoundedSemaphore(16)

    @property
    def name(self):
        return "ibmcloud"

    @staticmethod
    def region_list() -> List[str]:
        return REGIONS

    def setup_global(self, iam_name: str = "skyplane_gateway", attach_policy_arn: Optional[str] = None):
        # Not sure this should execute something. We will create VPC per region
        pass

    def setup_region(self, region: str):
        # set up VPC per region? With net, subnets, floating ip, etc. ?
        ibm_vpc_config = {
            "ibm": {"iam_api_key": self.auth.iam_api_key, "user_agent": self.auth.user_agent},
            "ibm_gen2": {"region": region, "resource_group_id": self.auth.ibmcloud_resource_group_id},
        }
        load_config(ibm_vpc_config)
        ibm_vpc_backend = IBMVPCBackend(ibm_vpc_config["ibm_gen2"])
        ibm_vpc_backend.init()
        self.regions_vpc[region] = ibm_vpc_backend

    def teardown_region(self, region):
        if region in self.regions_vpc:
            self.regions_vpc[region].clean(all=True)

    def teardown_global(self):
        for region in self.regions_vpc:
            self.regions_vpc[region].clean(all=True)

    def add_ips_to_security_group(self, cos_region: str, ips: Optional[List[str]] = None):
        return self.regions_vpc[cos_region].add_ips_to_security_group(ips)

    def remove_ips_from_security_group(self, cos_region: str, ips: List[str]):
        pass

    @imports.inject("botocore.exceptions", pip_extra="ibmcloud")
    def provision_instance(
        exceptions,
        self,
        region: str,
        instance_class: str,
        zone_name: Optional[str] = None,
        name: Optional[str] = None,
        tags={"skyplane": "true"},
    ) -> IBMCloudServer:
        # provision VM in the region

        tags["node-type"] = "master"
        tags["node-name"] = "skyplane-master"
        instance_id, vsi = self.regions_vpc[region].create_vpc_instance()
        return IBMCloudServer(self.regions_vpc[region], f"ibmcloud:{region}", instance_id, vsi)



import threading
from datetime import datetime
from datetime import datetime

import urllib3
from typing import TYPE_CHECKING, Dict, List, Optional

from skyplane import compute
from skyplane.api.tracker import TransferProgressTracker
from skyplane.api.transfer_job import CopyJob, SyncJob, TransferJob
from skyplane.api.config import TransferConfig

from skyplane.planner.planner import MulticastDirectPlanner, DirectPlannerSourceOneSided, DirectPlannerDestOneSided
from skyplane.planner.topology import TopologyPlanGateway
from skyplane.utils import logger
from skyplane.utils.definitions import tmp_log_dir

from skyplane.api.dataplane import Dataplane

if TYPE_CHECKING:
    from skyplane.api.provisioner import Provisioner


class Pipeline:
    """A pipeline object stores and executes a set of transfer jobs."""

    def __init__(
        self,
        clientid: str,
        provisioner: "Provisioner",
        transfer_config: TransferConfig,
        # cloud_regions: dict,
        max_instances: Optional[int] = 1,
        n_connections: Optional[int] = 64,
        planning_algorithm: Optional[str] = "direct",
        debug: Optional[bool] = False,
    ):
        """
        :param clientid: the uuid of the local host to create the dataplane
        :type clientid: str
        :param provisioner: the provisioner to launch the VMs
        :type provisioner: Provisioner
        :param transfer_config: the configuration during the transfer
        :type transfer_config: TransferConfig
        """
        self.clientid = clientid
        # self.cloud_regions = cloud_regions
        # TODO: set max instances with VM CPU limits and/or config
        self.max_instances = max_instances
        self.n_connections = n_connections
        self.provisioner = provisioner
        self.transfer_config = transfer_config
        self.http_pool = urllib3.PoolManager(retries=urllib3.Retry(total=3))
        self.provisioning_lock = threading.Lock()
        self.provisioned = False
        self.transfer_dir = tmp_log_dir / "transfer_logs" / datetime.now().strftime("%Y%m%d_%H%M%S")
        self.transfer_dir.mkdir(exist_ok=True, parents=True)

        # dataplane
        self.dataplane = None

        # planner
        self.planning_algorithm = planning_algorithm
        if self.planning_algorithm == "direct":
            self.planner = MulticastDirectPlanner(self.max_instances, self.n_connections, self.transfer_config)
        elif self.planning_algorithm == "src_one_sided":
            self.planner = DirectPlannerSourceOneSided(self.max_instances, self.n_connections, self.transfer_config)
        elif self.planning_algorithm == "dst_one_sided":
            self.planner = DirectPlannerDestOneSided(self.max_instances, self.n_connections, self.transfer_config)
        else:
            raise ValueError(f"No such planning algorithm {planning_algorithm}")

        # transfer logs
        self.transfer_dir = tmp_log_dir / "transfer_logs" / datetime.now().strftime("%Y%m%d_%H%M%S")
        self.transfer_dir.mkdir(exist_ok=True, parents=True)
        self.debug = debug

        # pending tracker tasks
        self.jobs_to_dispatch: List[TransferJob] = []
        self.pending_transfers: List[TransferProgressTracker] = []
        self.bound_nodes: Dict[TopologyPlanGateway, compute.Server] = {}

    def create_dataplane(self, debug):
        # create plan from set of jobs scheduled
        topo = self.planner.plan(self.jobs_to_dispatch)

        # create dataplane from plan
        dp = Dataplane(self.clientid, topo, self.provisioner, self.transfer_config, str(self.transfer_dir), debug=debug)
        return dp

    def start(self, debug=False, progress=False):
        ## create plan from set of jobs scheduled
        # topo = self.planner.plan(self.jobs_to_dispatch)

        ## create dataplane from plan
        # dp = Dataplane(self.clientid, topo, self.provisioner, self.transfer_config, self.transfer_dir, debug=debug)
        dp = self.create_dataplane(debug)
        try:
            dp.provision(spinner=True)
            if progress:
                from skyplane.cli.impl.progress_bar import ProgressBarTransferHook

                tracker = dp.run_async(self.jobs_to_dispatch, hooks=ProgressBarTransferHook(dp.topology.dest_region_tags))
            else:
                tracker = dp.run_async(self.jobs_to_dispatch)

            # wait for job to finish
            tracker.join()

            # copy gateway logs
            if debug:
                dp.copy_gateway_logs()
        except Exception:
            dp.copy_gateway_logs()
        dp.deprovision(spinner=True)
        return dp

    def start_async(self, debug=False):
        dp = self.create_dataplane(debug)
        try:
            dp.provision(spinner=False)
            tracker = dp.run_async(self.jobs_to_dispatch)
            if debug:
                dp.copy_gateway_logs()
            return tracker
        except Exception:
            dp.copy_gateway_logs()
            return

    def queue_copy(
        self,
        src: str,
        dst: str or List[str],
        recursive: bool = False,
    ) -> str:
        """
        Add a copy job to job list.
        Return the uuid of the job.

        :param src: source prefix to copy from
        :type src: str
        :param dst: the destination of the transfer
        :type dst: str
        :param recursive: if true, will copy objects at folder prefix recursively (default: False)
        :type recursive: bool
        """
        if isinstance(dst, str):
            dst = [dst]
        job = CopyJob(src, dst, recursive, requester_pays=self.transfer_config.requester_pays)
        logger.fs.debug(f"[SkyplaneClient] Queued copy job {job}")
        self.jobs_to_dispatch.append(job)
        return job.uuid

    def queue_sync(
        self,
        src: str,
        dst: str or List[str],
    ) -> str:
        """
        Add a sync job to job list.
        Return the uuid of the job.

        :param src: Source prefix to copy from
        :type src: str
        :param dst: The destination of the transfer
        :type dst: str
        :param recursive: If true, will copy objects at folder prefix recursively (default: False)
        :type recursive: bool
        """
        if isinstance(dst, str):
            dst = [dst]
        job = SyncJob(src, dst, requester_pays=self.transfer_config.requester_pays)
        logger.fs.debug(f"[SkyplaneClient] Queued sync job {job}")
        self.jobs_to_dispatch.append(job)
        return job.uuid

    def estimate_total_cost(self):
        """Estimate total cost of queued jobs"""
        total_size = 0
        for job in self.jobs_to_dispatch:
            total_size += job.size_gb()

        # get planner topology
        topo = self.planner.plan(self.jobs_to_dispatch)

        # return size
        return total_size * topo.cost_per_gb



import base64
import hashlib
import os
from functools import lru_cache

from typing import Any, Iterator, List, Optional, Tuple

from skyplane import exceptions, compute
from skyplane.exceptions import NoSuchObjectException
from skyplane.obj_store.object_store_interface import ObjectStoreInterface, ObjectStoreObject
from skyplane.config_paths import cloud_config
from skyplane.utils import logger, imports
from skyplane.utils.generator import batch_generator


class S3Object(ObjectStoreObject):
    def full_path(self):
        return f"s3://{self.bucket}/{self.key}"


class S3Interface(ObjectStoreInterface):
    def __init__(self, bucket_name: str):
        self.auth = compute.AWSAuthentication()
        self.requester_pays = False
        self.bucket_name = bucket_name
        self._cached_s3_clients = {}

    @property
    def provider(self):
        return "aws"

    def path(self):
        return f"s3://{self.bucket_name}"

    @property
    @lru_cache(maxsize=1)
    def aws_region(self):
        s3_client = self.auth.get_boto3_client("s3")
        default_region = cloud_config.get_flag("aws_default_region")
        try:
            # None means us-east-1 for legacy reasons
            region = s3_client.get_bucket_location(Bucket=self.bucket_name).get("LocationConstraint", "us-east-1")
            return region if region is not None else default_region
        except Exception as e:
            if "An error occurred (AccessDenied) when calling the GetBucketLocation operation" in str(e):
                logger.warning(f"Bucket location {self.bucket_name} is not public. Assuming region is {default_region}")
                return default_region
            elif "An error occurred (InvalidAccessKeyId) when calling" in str(e):
                logger.warning(f"Invalid AWS credentials. Check to make sure credentials configured properly.")
                raise exceptions.PermissionsException(f"Invalid AWS credentials for accessing bucket {self.bucket_name}")
            else:
                logger.warning(f"Specified bucket {self.bucket_name} does not exist, got AWS error: {e}")
                raise exceptions.MissingBucketException(f"S3 bucket {self.bucket_name} does not exist") from e

    def region_tag(self):
        return "aws:" + self.aws_region

    def bucket(self) -> str:
        return self.bucket_name

    def set_requester_bool(self, requester: bool):
        self.requester_pays = requester

    def _s3_client(self, region=None):
        region = region if region is not None else self.aws_region
        if region not in self._cached_s3_clients:
            self._cached_s3_clients[region] = self.auth.get_boto3_client("s3", region)
        return self._cached_s3_clients[region]

    @imports.inject("botocore.exceptions", pip_extra="aws")
    def bucket_exists(botocore_exceptions, self, region=None):
        if region is None:  # use current bucket region is available
            try:
                region = self.aws_region
            except exceptions.MissingBucketException:
                region = "us-east-1"

        s3_client = self._s3_client(region)
        try:
            requester_pays = {"RequestPayer": "requester"} if self.requester_pays else {}
            s3_client.list_objects_v2(Bucket=self.bucket_name, MaxKeys=1, **requester_pays)
            return True
        except botocore_exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchBucket" or e.response["Error"]["Code"] == "AccessDenied":
                return False
            raise

    def create_bucket(self, aws_region):
        s3_client = self._s3_client(aws_region)
        if not self.bucket_exists(aws_region):
            if aws_region == "us-east-1":
                s3_client.create_bucket(Bucket=self.bucket_name)
            else:
                s3_client.create_bucket(Bucket=self.bucket_name, CreateBucketConfiguration={"LocationConstraint": aws_region})
        else:
            logger.warning(f"Bucket {self.bucket} in region {aws_region} already exists")

    def delete_bucket(self):
        # delete 1000 keys at a time
        for batch in batch_generator(self.list_objects(), 1000):
            self.delete_objects([obj.key for obj in batch])
        assert len(list(self.list_objects())) == 0, f"Bucket not empty after deleting all keys {list(self.list_objects())}"
        # delete bucket
        self._s3_client().delete_bucket(Bucket=self.bucket_name)

    def list_objects(self, prefix="", region=None) -> Iterator[S3Object]:
        paginator = self._s3_client(region).get_paginator("list_objects_v2")
        requester_pays = {"RequestPayer": "requester"} if self.requester_pays else {}
        page_iterator = paginator.paginate(Bucket=self.bucket_name, Prefix=prefix, **requester_pays)
        for page in page_iterator:
            objs = []
            for obj in page.get("Contents", []):
                objs.append(
                    S3Object(
                        obj["Key"],
                        provider=self.provider,
                        bucket=self.bucket(),
                        size=obj["Size"],
                        last_modified=obj["LastModified"],
                        mime_type=obj.get("ContentType"),
                    )
                )
            yield from objs

    def delete_objects(self, keys: List[str]):
        s3_client = self._s3_client()
        while keys:
            batch, keys = keys[:1000], keys[1000:]  # take up to 1000 keys at a time
            s3_client.delete_objects(Bucket=self.bucket_name, Delete={"Objects": [{"Key": k} for k in batch]})

    @lru_cache(maxsize=1024)
    @imports.inject("botocore.exceptions", pip_extra="aws")
    def get_obj_metadata(botocore_exceptions, self, obj_name):
        s3_client = self._s3_client()
        try:
            return s3_client.head_object(Bucket=self.bucket_name, Key=str(obj_name))
        except botocore_exceptions.ClientError as e:
            raise NoSuchObjectException(f"Object {obj_name} does not exist, or you do not have permission to access it") from e

    def get_obj_size(self, obj_name):
        return self.get_obj_metadata(obj_name)["ContentLength"]

    def get_obj_last_modified(self, obj_name):
        return self.get_obj_metadata(obj_name)["LastModified"]

    def get_obj_mime_type(self, obj_name):
        return self.get_obj_metadata(obj_name)["ContentType"]

    def exists(self, obj_name):
        try:
            self.get_obj_metadata(obj_name)
            return True
        except NoSuchObjectException:
            return False

    def download_object(
        self,
        src_object_name,
        dst_file_path,
        offset_bytes=None,
        size_bytes=None,
        write_at_offset=False,
        generate_md5=False,
        write_block_size=2**16,
    ) -> Tuple[Optional[str], Optional[bytes]]:
        src_object_name, dst_file_path = str(src_object_name), str(dst_file_path)

        s3_client = self._s3_client()
        assert len(src_object_name) > 0, f"Source object name must be non-empty: '{src_object_name}'"
        args = {"Bucket": self.bucket_name, "Key": src_object_name}
        assert not (offset_bytes and not size_bytes), f"Cannot specify {offset_bytes} without {size_bytes}"
        if offset_bytes is not None and size_bytes is not None:
            args["Range"] = f"bytes={offset_bytes}-{offset_bytes + size_bytes - 1}"
        if self.requester_pays:
            args["RequestPayer"] = "requester"
        response = s3_client.get_object(**args)

        # write response data
        if not os.path.exists(dst_file_path):
            open(dst_file_path, "a").close()
        if generate_md5:
            m = hashlib.md5()
        with open(dst_file_path, "wb+" if write_at_offset else "wb") as f:
            f.seek(offset_bytes if write_at_offset else 0)
            b = response["Body"].read(write_block_size)
            while b:
                if generate_md5:
                    m.update(b)
                f.write(b)
                b = response["Body"].read(write_block_size)
        response["Body"].close()
        md5 = m.digest() if generate_md5 else None
        mime_type = response["ContentType"]
        return mime_type, md5

    @imports.inject("botocore.exceptions", pip_extra="aws")
    def upload_object(
        botocore_exceptions, self, src_file_path, dst_object_name, part_number=None, upload_id=None, check_md5=None, mime_type=None
    ):
        dst_object_name, src_file_path = str(dst_object_name), str(src_file_path)
        s3_client = self._s3_client()
        assert len(dst_object_name) > 0, f"Destination object name must be non-empty: '{dst_object_name}'"
        b64_md5sum = base64.b64encode(check_md5).decode("utf-8") if check_md5 else None
        checksum_args = dict(ContentMD5=b64_md5sum) if b64_md5sum else dict()

        try:
            with open(src_file_path, "rb") as f:
                if upload_id:
                    s3_client.upload_part(
                        Body=f,
                        Key=dst_object_name,
                        Bucket=self.bucket_name,
                        PartNumber=part_number,
                        UploadId=upload_id.strip(),  # TODO: figure out why whitespace gets added,
                        **checksum_args,
                    )
                else:
                    mime_args = dict(ContentType=mime_type) if mime_type else dict()
                    s3_client.put_object(Body=f, Key=dst_object_name, Bucket=self.bucket_name, **checksum_args, **mime_args)
        except botocore_exceptions.ClientError as e:
            # catch MD5 mismatch error and raise appropriate exception
            if "Error" in e.response and "Code" in e.response["Error"] and e.response["Error"]["Code"] == "InvalidDigest":
                raise exceptions.ChecksumMismatchException(f"Checksum mismatch for object {dst_object_name}") from e
            raise

    def initiate_multipart_upload(self, dst_object_name: str, mime_type: Optional[str] = None) -> str:
        client = self._s3_client()
        assert len(dst_object_name) > 0, f"Destination object name must be non-empty: '{dst_object_name}'"
        response = client.create_multipart_upload(
            Bucket=self.bucket_name, Key=dst_object_name, **(dict(ContentType=mime_type) if mime_type else dict())
        )
        if "UploadId" in response:
            return response["UploadId"]
        else:
            raise exceptions.SkyplaneException(f"Failed to initiate multipart upload for {dst_object_name}: {response}")

    def complete_multipart_upload(self, dst_object_name, upload_id, metadata: Optional[Any] = None):
        s3_client = self._s3_client()
        all_parts = []
        while True:
            response = s3_client.list_parts(
                Bucket=self.bucket_name, Key=dst_object_name, MaxParts=100, UploadId=upload_id, PartNumberMarker=len(all_parts)
            )
            if "Parts" not in response:
                break
            else:
                if len(response["Parts"]) == 0:
                    break
                all_parts += response["Parts"]
        all_parts = sorted(all_parts, key=lambda d: d["PartNumber"])
        response = s3_client.complete_multipart_upload(
            UploadId=upload_id,
            Bucket=self.bucket_name,
            Key=dst_object_name,
            MultipartUpload={"Parts": [{"PartNumber": p["PartNumber"], "ETag": p["ETag"]} for p in all_parts]},
        )
        assert "ETag" in response, f"Failed to complete multipart upload for {dst_object_name}: {response}"
